indexing

   description: "Class that covers a portable directory. Main use is to be %
   %able to loop through a single directory."

   author: "Berend de Boer, Marcio Marchini"
   date: "$Date: 2001/12/27 $";
   revision: "$Revision: #1 $"


deferred class

   ABSTRACT_DIRECTORY


inherit

   STDC_BASE

   EPX_FACTORY
      export
         {NONE} all
      end


feature -- creation

   make (a_directory_name: STRING) is
      do
         create directory_name.make_from_string (a_directory_name)
         if my_item = Void then
            my_item := sh.make_with_capacity (max_filename_length)
         end
      end


feature -- access

   close is
         -- close directory entry (save resources now, don't wait for
         -- garbage collection). If you call start it will automatically
         -- reopen.
         -- if forth ends with exhausted, close is automatically called.
      require
         not_closed: is_open
      do
         safe_call (abstract_closedir (dirp))
         dirp := default_pointer
      ensure
         closed: not is_open
      end

   start is
      -- start directory traversal
      do
         exhausted := False
         subdirectory := Void
         -- remove create when ELKS2001 is supported by everyone
         my_item.wipe_out
         if dirp = default_pointer then
            dirp := abstract_opendir (sh.string_to_pointer(directory_name))
            if dirp = default_pointer then
               raise_posix_error
            end
         else
            abstract_rewinddir (dirp)
         end
         forth
         first := True
      end

   forth is
      require
         opened: is_open
         not_exhausted: not exhausted
      local
         stop: BOOLEAN
         optional_status: ABSTRACT_STATUS
      do
         first := False
         if filter = Void then
            forth_recursive
         else
            from
               stop := False
            until
               stop
            loop
               forth_recursive
               if not exhausted and then filter.require_status then
                  optional_status := status
               end
               stop := exhausted or else filter.is_valid (optional_status,
                                                          full_name)
            end
         end
      ensure
         no_handle_leak: not is_open implies dirp = default_pointer
      end


feature -- influence browsing

   filter: ABSTRACT_PATH_FILTER

   set_filter (a_filter: ABSTRACT_PATH_FILTER) is
      do
         filter := a_filter
      end

   set_extension_filter (extension: STRING) is
         -- create and set filter, `extension' is like ".e"
         -- `extension' shouldn't contain any wildcards
      local
         ext_filter: EXTENSION_FILTER
      do
         create ext_filter.make (extension)
         set_filter (ext_filter)
      end

   recursive: BOOLEAN
         -- enable depth first directory browsing

   set_recursive (enable: BOOLEAN) is
      do
         recursive := enable
      ensure
         recursive_set: recursive = enable
      end


feature -- public state

   directory_name: STDC_PATH

   first: BOOLEAN

   full_name: STDC_PATH is
         -- combination of directory_name and item
      require
         opened: is_open
      do
         if subdirectory = Void then
            create Result.make (directory_name.count +
                                1 +
                                item.count)
            Result.append_string (directory_name)
            Result.append_character (fs.path_separator)
            Result.append_string (item)
         else
            Result := subdirectory.full_name
         end
      end

   item: STRING is
         -- the current entry, i.e. the path name.
         -- even for recursive browsing, this entry is just the path
         -- name without any preceding sub directory name
      require
         opened: is_open
      do
         if subdirectory = Void then
            Result := my_item
         else
            Result := subdirectory.item
         end
      end

   status: ABSTRACT_STATUS is
      do
         if subdirectory = Void then
            if my_status = Void then
               my_status := fs.status (full_name)
            end
            Result := my_status
         else
            Result := subdirectory.status
         end
      end


feature -- status report

   exhausted: BOOLEAN
         -- no more entries in this directory

   is_open: BOOLEAN is
         -- True if directory is ready for traversal
      do
         Result := dirp /= default_pointer
      end


feature -- other directory related queries

   max_filename_length: INTEGER is
         -- maximum length of a file in this directory
      deferred
      end


feature -- annihilation

   dispose is
      do
         if is_open then
            close
         end
      end


feature {NONE} -- directory browsing primitives

   forth_recursive is
      -- next entry, depth-first recursive if needed
      require
         opened: is_open
         not_exhausted: not exhausted
      local
         special_dir: BOOLEAN
      do
         if subdirectory /= Void then
            subdirectory.forth
            if subdirectory.exhausted then
               subdirectory := Void
            end
         else
            forth_single
            if recursive and not exhausted then
               special_dir := item.is_equal (".") or else item.is_equal ("..")
               if not special_dir then
                  my_status := fs.status (full_name)
                  if my_status.is_directory then
                     subdirectory := fs.browse_directory (full_name)
                     subdirectory.set_recursive (True)
                     subdirectory.start
                     subdirectory.forth
                  end
               end
            end
         end
      end

   forth_single is
      -- go to next entry
      require
         opened: is_open
         not_exhausted: not exhausted
      local
         dirent: POINTER
      do
         errno.clear
         my_status := Void
         dirent := abstract_readdir (dirp)
         if dirent = default_pointer then
            if errno.is_not_ok then
               raise_posix_error
            else
               exhausted := True
               sh.clear (my_item)
               close
            end
         else
            sh.set_string_from_pointer(my_item, abstract_d_name (dirent))
         end
      end


feature {NONE} -- private state

   dirp: POINTER

   my_item: STRING
         -- the current entry, i.e. the path name

   my_status: ABSTRACT_STATUS
         -- optional status of current entry

   subdirectory: ABSTRACT_DIRECTORY


feature {NONE} -- abstract API

   abstract_closedir (a_dirp: POINTER): INTEGER is
         -- Ends directory read operation
      deferred
      end

   abstract_opendir (a_dirname: POINTER): POINTER is
         -- Opens a directory
      deferred
      end

   abstract_readdir (a_dirp: POINTER): POINTER is
         -- Reads a directory
      deferred
      end

   abstract_rewinddir (a_dirp: POINTER) is
         -- Resets the readdir pointer
      deferred
      end

   abstract_d_name (a_dirent: POINTER): POINTER is
      deferred
      end


invariant

   valid_directory_name: directory_name /= Void

end -- class ABSTRACT_DIRECTORY
