indexing

   description: "Class that covers Standard C time related routines. %
   %Assumes that ANSI C is POSIX compliant, so time is defined with %
   %base January 1, 1970. So dates before 1970 and after 2038 are not %
   %really supported...%
   %Time resolutation is one second."

   known_deficiencies: "1. There is no way to create a time based on UTC time%
   %fields yet."

   author: "Berend de Boer"
   date: "$Date: 2001/12/27 $"
   revision: "$Revision: #1 $"

class 

   STDC_TIME

inherit

   STDC_BASE
      redefine
         is_equal
      end

   COMPARABLE
      redefine
         is_equal
      end

   HASHABLE
      redefine 
         is_equal
      end

   CAPI_TIME
      redefine
         is_equal
      end


creation

   make_date,
   make_date_time,
   make_from_now,
   make_time,
   make_from_unix_time


feature -- creation

   make_date (a_year, a_month, a_day: INTEGER) is
         -- create a time according to this day, time 00:00:00
         -- date is assumed to be local date
      do
         make_date_time (a_year, a_month, a_day, 0, 0, 0)
      end

   make_date_time (a_year, a_month, a_day, a_hour, a_minute, a_second: INTEGER) is
         -- date is assumed to be local date
         -- we assume daylight saving time setting in effect is 
         -- available from system
      do
         assert_has_tm
         set_dst_to_current
         set_date_time (a_year, a_month, a_day, a_hour, a_minute, a_second)
      end
  
   make_from_now is
         -- make `value' equal to current unix time
         -- afterwards call to_local or to_utc to turn individual
         -- fields in local time or in utc time.
      do
         make_from_unix_time (posix_time)
      end

   make_time (a_hour, a_minute, a_second: INTEGER) is
         -- we assume daylight saving time setting in effect is 
         -- available from system
         -- day will be January 1, 1970
      do
         assert_has_tm
         make_date_time (minimum_year, 1, 1, a_hour, a_minute, a_second)
      end

   make_from_unix_time (a_value: INTEGER) is
      do
         assert_has_tm
         value := a_value
      end
   
   
feature -- make individual time fields valid
   
   to_local is
         -- switch time fields to local time
      local
         p: POINTER
      do
         p := posix_localtime(value)
         tm.memory_copy (p, 0, 0, posix_tm_size)
      end

   to_utc is
         -- switch time fields to utc time
      local
         p: POINTER
      do
         p := posix_gmtime(value)
         tm.memory_copy (p, 0, 0, posix_tm_size)
      end


feature -- manually set individual time fields   
   
   set_date (a_year, a_month, a_day: INTEGER) is
         -- set date part, time remains unchanged
      do
         posix_set_tm_date (tm.ptr, a_year - 1900, a_month - 1, a_day)
         value := posix_mktime (tm.ptr)
      end

   set_date_time (a_year, a_month, a_day, a_hour, a_minute, a_second: INTEGER) is
         -- we assume daylight saving time setting in effect (or not)
         -- has been set
      require
         valid_year: a_year >= minimum_year
         -- valid date/time: don't try to pass invalid values
      do
         posix_set_tm_date (tm.ptr, a_year - 1900, a_month - 1, a_day)
         posix_set_tm_time (tm.ptr, a_hour, a_minute, a_second)
         value := posix_mktime (tm.ptr)
         if value = -1 then
            raise_posix_error
         end
      end

   set_dst_to_current is
         -- let system figure out if daylight saving time is in effect
      do
         posix_set_tm_dst_not_available (tm.ptr)
      end
   
   set_dst_to_none is
         -- daylight saving time is not in effect
      do
         posix_set_tm_dst_none (tm.ptr)
      end
   
   set_time (a_hour, a_minute, a_second: INTEGER) is
         -- set time part, date remains unchanged
      do
         posix_set_tm_time (tm.ptr, a_hour, a_minute, a_second)
         value := posix_mktime (tm.ptr)
      end
   
   
feature -- individual time fields, need call to `to_local' or `to_utc'

   year: INTEGER is
      do
         Result := 1900 + posix_tm_year (tm.ptr)
      end
   
   month: INTEGER is
      do
         Result := posix_tm_mon (tm.ptr) + 1
      end

   day: INTEGER is
         -- Day of the month
      do
         Result := posix_tm_mday (tm.ptr)
      end

   weekday: INTEGER is
         -- Days since Sunday
      do
         Result := posix_tm_wday (tm.ptr)
      end

   day_of_year: INTEGER is
         -- Days since January 1st
      do
         Result := posix_tm_yday (tm.ptr)
      end
   
   hour: INTEGER is
      do
         Result := posix_tm_hour (tm.ptr)
      end
   
   minute: INTEGER is
      do
         Result := posix_tm_min (tm.ptr)
      end
   
   second: INTEGER is
      do
         Result := posix_tm_sec (tm.ptr)
      end
  

feature -- time as string   

   short_weekday_name: STRING is
         -- Abbreviated weekday name
      do
         Result := format ("%%a")
      end
   
   weekday_name: STRING is
         -- Full weekday name
      do
         Result := format ("%%A")
      end

   short_month_name: STRING is
         -- Abbreviated month name
      do
         Result := format ("%%b")
      end
   
   month_name: STRING is
         -- Full month name
      do
         Result := format ("%%B")
      end
   
   format (format_str: STRING): STRING is
      local
         p: STDC_BUFFER
         r: INTEGER
      do
         create p.allocate (128)
         r := posix_strftime (p.ptr, 128, sh.string_to_pointer (format_str), tm.ptr)
         if r = 0 then
            raise_posix_error
         end
         Result := sh.pointer_to_string (p.ptr)
      end

   default_format: STRING is
         -- Returns a string of the form "Mon Apr 17 21:49:20 2000"
      do
         Result := sh.pointer_to_string (posix_asctime (tm.ptr))
         Result.remove (Result.count)
      end

   local_date_string: STRING is
         -- Return date part in format local to current country
      do
         Result := format ("%%x")
      end

   local_time_string: STRING is
         -- Return time part in format local to current country
      do
         Result := format ("%%X")
      end


feature -- date calculations
   
   is_equal(other: like Current): BOOLEAN is
      do
         Result := other.value = value
      end

   infix "-" (other: like Current): like Current is
         -- Creates a new time which is the difference between
         -- `Current' and `Other'
      local
         diff: DOUBLE
      do
         diff := posix_difftime(value, other.value)
         create Result.make_from_unix_time (diff.truncated_to_integer)
      end

   infix "<" (other: like Current): BOOLEAN is
         -- Is current object less than `other'?
      local
         diff: DOUBLE
      do
         diff := posix_difftime(value, other.value)
         Result := diff < 0
      end
   
   
feature -- state

   hash_code: INTEGER is
      do
         Result := value
      end
   
   value: INTEGER
         -- time in seconds from January 1, 1970
         -- perhaps since 1980 for Windows systems


feature -- non POSIX, but Gates specific stuff
   
   minimum_year: INTEGER is
         -- returns the minimum year for the current platform
         -- for POSIX is 1970, for Windows is 1980
      once
         if is_windows then
            Result := 1980
         else
            Result := 1970
         end
      end
   
   
feature {NONE} -- state

   tm: STDC_BUFFER
   
   assert_has_tm is
      do
         if tm = Void then
            create tm.allocate_and_clear (posix_tm_size)
         end
      end


feature {NONE}

   is_windows: BOOLEAN is
      external "C"
      end


invariant
   
   valid_tm_struct: tm /= Void

end -- class STDC_TIME
